#include <cstdio>
#include <sys/time.h> /* gettimeofday, setitimer */
#include <termio.h>   /* termio */

#include <sys/resource.h> /* PRIO_* */

#include <csignal>

#include <algorithm>
#include <list>

/* Captures video from a X window.
 * Requires the SHM extension to work on tolerable speeds.
 *
 * Copyright (C) 1992,2004 Bisqwit (http://iki.fi/bisqwit/)
 *
 * TODO: Create a standalone program without all parameters being
 *       hardcoded into the program.
 *
 * To compile: gcc -o xgetimage xgetimage.c -L/usr/X11R6/lib -lX11 -lXext -lXmu
 *
 */

#include "settings.hh"
#include "buffer.hh"
#include "msg.hh"
#include "x.hh"

#define BGR24 0x42475218  // BGR24 fourcc
#define BGR16 0x42475210  // BGR16 fourcc
#define BGR15 0x4247520F  // BGR15 fourcc

typedef unsigned long long UINT64;
static UINT64 getutime(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000LL) + tv.tv_usec;
}

static unsigned frames_total    = 0;
static unsigned frames_thisfile = 0;
static UINT64 start_time;

static std::list<FILE *> finished_encoders;
static bool empty = false;
static FILE *avifp= NULL;

static FILE *OpenAvi()
{
    static unsigned avinumber = 0;
    char CmdLine[512];
    char avifn[64];
    sprintf(avifn, "test%d.avi", avinumber++);

    string options = "vcodec=mpeg4";
    //options += ":qpel";
    //options += ":vrc_eq=1";
    options += ":dia=-3:predia=-3:preme=2";
    options += ":v4mv";
    options += ":vb_strategy=1:vqscale=1";
    //options += ":cmp=1:precmp=0:subcmp=0";
    //options += ":keyint=30";

    sprintf(CmdLine,
        "mencoder - -o %s -ovc lavc"
        " -lavcopts %s:vbitrate=%u"
        //:vrc_maxrate=%u:vrc_minrate=%u"
        " >& /dev/null"
        //"cat > %s"
        ,
            avifn,
            options.c_str(),
            BITRATE
          //  BITRATE,BITRATE
           );

    MessageCreatingFile(avifn);

    frames_thisfile = 0;

    setpriority(PRIO_PGRP, 0, 0);
    return popen(CmdLine, "w");
}

static void InitAvi()
{
    start_time   = getutime();
    avifp        = OpenAvi();
    frames_total = 0;
    empty        = true;
}

void BufferEmptied(FILE *fp)
{
    //fprintf(stderr, "\nBuffer %X emptied\n", fp);
    for(std::list<FILE*>::iterator
        j, i = finished_encoders.begin();
           i != finished_encoders.end();
        i = j)
    {
        j = i; ++j;
        
        if(fp == *i)
        {
            //fprintf(stderr, "\nFound and closed.\n");
            pclose(*i);
            finished_encoders.erase(i);
            return;
        }
    }
    if(fp == avifp) empty = true; 
}

static void DoneAvi()
{
    FlushBuffers();
    
    for(std::list<FILE*>::iterator
        i = finished_encoders.begin();
        i != finished_encoders.end();
        ++i)
    {
        pclose(*i);
    }
    
    finished_encoders.clear();
    avifp   = NULL;
    empty   = true;
}
static void NextAvi()
{
    if(empty)
        pclose(avifp);
    else
        finished_encoders.push_back(avifp);
    avifp = OpenAvi();
    empty = true;
}

static double GetVideoExcess()
{
    UINT64 time_now = getutime();
    double videopos = frames_total / (double)FPS;
    double want_pos = (time_now - start_time) / 1e6;
    return videopos - want_pos;
}

#define u32(n) (n)&255,((n)>>8)&255,((n)>>16)&255,((n)>>24)&255
#define u16(n) (n)&255,((n)>>8)&255
#define s4(s) s[0],s[1],s[2],s[3]

static void SendFrame(unsigned char *vidbuf)
{
    const unsigned fourcc = BGR16;
    const unsigned framesize = width*height*2;
    
    const unsigned char header[] = { s4("00dc"), u32(framesize) };
    
    BufferType thisframe(sizeof(header) + framesize);
    
    std::copy(header, header + sizeof header,
              &thisframe[0]);
    
    std::copy(vidbuf, vidbuf + framesize,
              &thisframe[sizeof header]);
    
    
    if(FRAMES_PER_FILE > 0
    && frames_thisfile >= FRAMES_PER_FILE)
    {
        NextAvi();
    }
    
    if(!frames_thisfile)
    {
        const unsigned nframes = 0; //unknown
        const unsigned scale = 1000000;
        const unsigned scaled_fps = (unsigned)(scale * FPS);
        
        const unsigned SIZE_strh_vids = 4 + 4*2 + 2*2 + 8*4 + 2*4;
        const unsigned SIZE_strf_vids = 4*3 + 2*2 + 4*6;
        const unsigned SIZE_strl = 4+ 4+(4+SIZE_strh_vids) + 4+(4+SIZE_strf_vids);
        const unsigned SIZE_avih = 4*12;
        const unsigned SIZE_hdrl = 4+4+ (4+SIZE_avih) + 4 + (4+SIZE_strl);
        const unsigned SIZE_movi = 4 + nframes*(4+4+framesize);
        const unsigned SIZE_avi = 4+4+ (4+SIZE_hdrl) + 4 + (4+SIZE_movi);
        
        const unsigned char AVIheader[] =
        {
            s4("RIFF"),
            u32(SIZE_avi),
            s4("AVI "),
            s4("LIST"),
            u32(SIZE_hdrl),
             s4("hdrl"),
             
             s4("avih"),
             u32(SIZE_avih),
              u32(0),
              u32(0),
              u32(0),
              u32(0),
              u32(nframes),
              u32(0),
              u32(1), // one stream
              u32(0),
              u32(0),
              u32(0),
              u32(0),
              u32(0),
             
             s4("LIST"),
             
             u32(SIZE_strl),
              s4("strl"),
              
               s4("strh"),
               u32(SIZE_strh_vids),
                s4("vids"),
                u32(0),
                u32(0),
                u16(0),
                u16(0),
                u32(0),
                u32(scale),
                u32(scaled_fps),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
                u16(0),
                u16(0),
                u16(0),
                u16(0),
               
               s4("strf"),
               u32(SIZE_strf_vids),
                u32(0),
                u32(width),
                u32(height),
                u16(0),
                u16(0),
                u32(fourcc),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
            
            s4("LIST"),
            u32(SIZE_movi),
             s4("movi")
        };
        
        BufferType aviheader(AVIheader, AVIheader + sizeof(AVIheader));
        AddWriteTask(avifp, aviheader);
    }
    
    empty = false;
    AddWriteTask(avifp, thisframe);
    
    ++frames_total;
    ++frames_thisfile;
}

static void HandleImage(unsigned char *vidbuf)
{
    unsigned frames_now = 0;
    
    // Set a high threshold for the first frame
    // to make it unlikely to discard frames
    double threshold = 0.333 / FPS;
    
    for(;;)
    {
        double excess = GetVideoExcess();
        if(excess > threshold) break; /* discard frame, we've had enough already */
        threshold = 0;

        SendFrame(vidbuf);
        ++frames_now;
    }
    
    if(!frames_now)
        MessageDiscardFrame();
    else
        MessageCapturedFrame(frames_now);
}

static termio back;
static void InitKB()
{
    struct termio term;
    ioctl(0, TCGETA, &term);
    ioctl(0, TCGETA, &back);
    term.c_lflag &= ~(ECHO | ICANON);
    term.c_cc[VMIN] = 0; 
    ioctl(0, TCSETA, &term);
}
static void DoneKB()
{
    ioctl(0, TCSETA, &back);
}

static void SetTimeVal(struct timeval& t, double value)
{
    t.tv_sec  = (unsigned) value;
    t.tv_usec = (unsigned) ((value - t.tv_sec) * 1e6);
}

static void CaptureFrame(int) { /* nothing to do */ }
static void InitializeTimers()
{
    struct itimerval tr;
    SetTimeVal(tr.it_interval, 1.0 / FPS);
    
    double next = GetVideoExcess() * DELAY_MULTIPLIER;

    // Set a minimum to ensure the timer won't be disabled
    if(next <= 1.4e-6) next = 1.4e-6;

    SetTimeVal(tr.it_value,    next);

    setpriority(PRIO_PROCESS, 0, -20);
    setitimer(ITIMER_REAL, &tr, NULL);
    
    /*
    *   FIXME: SEE ALSO "man 3p timer_create", "man 3p timer_getoverrun"
    *
    **/
}
static void EndTimers()
{
    struct itimerval tr;
    SetTimeVal(tr.it_value,    0);
    SetTimeVal(tr.it_interval, 0);
    setitimer(ITIMER_REAL, &tr, NULL);
    signal(SIGALRM, SIG_DFL);
}

int main(int argc, const char*const *argv)
{
    XInit();
    
    // Get 1 image to ensure there are no
    // first-initialization-delays waiting for us.
    GetImage();
    DeleteImage();
    
    InitKB();
    
    signal(SIGALRM, CaptureFrame);

    InitAvi();
    
    InitializeTimers();
    
    while(!BufferWaitInput())
    {
        unsigned char *data = GetImage();
        if(data)
        {
            HandleImage(data);
            MessageFrameCounter(frames_total);
        }
        DeleteImage();
        
        InitializeTimers();
    }
    
    EndTimers();
    
    DoneKB();
    DoneAvi();
    
    MessageEnd();
    
    XDone();
    
    return 0;
}
